#!/usr/bin/perl -w
use strict;

# BAK_visual.pl
#
#  Copyright 2007 University of Pittsburgh.
#  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
#
#  This file is part of BAK (Branch-and-bound Analysis Kit).
#
#  The contents of this file are subject to the Common Public License
#  1.0.  (the "License"); you may not use this file except in
#  compliance with the License. You should have received a copy of
#  the Common Public License along with STOP.
#
#  Software distributed under the License is distributed on an "AS
#  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
#  implied. See the License for the specific language governing
#  rights and limitations under the License.
#
#  Alternatively, the contents of this file may be used under the
#  terms of the GNU General Public License Version 2 or later (the
#  "GPL"), in which case the provisions of the GPL are applicable
#  instead of those above. If you wish to allow use of your version
#  of this file only under the terms of the GPL, and not to allow
#  others to use your version of this file under the terms of the
#  CPL, indicate your decision by deleting the provisions above and
#  replace them with the notice and other provisions required by the
#  GPL. If you do not delete the provisions above, a recipient may
#  use your version of this file under the terms of either the CPL or
#  the GPL.
#

################ OPEN INPUT FILE AND SET NUMBER OF IMAGES ####################

use File::Spec;  # just for finding the directory the executable is in
use Getopt::Long; # option parsing
use constant EPS => 10**(-10);


# Options
my $help = "";      # show help message
my $interval = 10;
my $timelimit = -1; # must be a negative number
my $label = "";     # text added to every image
my $check_fathom = "";   # whether to do our own check for fathomed nodes.
                         # This simply omits nodes that should have been
                         # fathomed from the relevant images.

my $do_histogram = "";   # whether to print histograms
my $do_scatterplot = ""; # whether to print scatterplots
my $do_path = "";        # whether to print incumbent paths 
my $do_tree = "";        # whether to print trees, adjusted horizontal
my $do_fixedtree = "";   # whether to print trees, fixed horizontal
my $do_predictions = ""; # whether to print predictions

my $do_all = "";         # activates everything

my $result = GetOptions("help" => \$help,
			"interval=i" => \$interval,  # time between images
			"timelimit=i" => \$timelimit,
			"label=s" => \$label,   # text to add on all images
			"fathom!" => \$check_fathom,
			"histogram!" => \$do_histogram,
			"scatterplot!" => \$do_scatterplot,
			"path!" => \$do_path,
			"tree!" => \$do_tree,
			"fixedtree!" => \$do_fixedtree,
			"predictions!" => \$do_predictions,
			"all" => \$do_all
			);

if (!$result || $help) { 
    print_usage(); 
    exit(0);
}

if ($do_all) { 
    # activate everything unless it had been specifically disabled
    if ($do_histogram eq "") { $do_histogram = 1; }
    if ($do_scatterplot eq "") { $do_scatterplot = 1; }
    if ($do_path eq "") { $do_path = 1;}
    if ($do_tree eq "") { $do_tree = 1; }
    if ($do_fixedtree eq "") { $do_fixedtree = 1; }
    if ($do_predictions eq "") { $do_predictions = 1; }
}


# There should be exactly one argument other than the options
if ($#ARGV != 0) { 
    print_usage(); 
    exit(0);
}
my $infile = $ARGV[0];

# Record the directories for calling other scripts
my $directories;
(undef,$directories,undef) = File::Spec->splitpath( $0 );


print "problem: $infile\n";


################ OPEN OUTPUT FILE TO STORE SCRIPTFILE NAMES ##################

# This file enables us to plot all script file with a single command
my $plotfiles = "scriptfiles.dat";
open (PLOTFILES, "> $plotfiles") or die "Couldn't open file $plotfiles\n";
# array of plot filenames
my @plotfilenames;  

#################### GLOBAL VARIABLE DEFINITIONS ##############################


my $input;
my $opt_dir = 0;
my $incumbent;
my $past_incumbent;
my $incumbent_parent;
my $nlines = 0;
my %nodes;
my @keylist;
my $minobj;
my $maxobj;
my $maxiicount = 0;
my $maxiisum = 0;
my $linecounter = 0;
my $imagenum = 0;
my $imagename = 0;
my $rootID;
my $ref_counter;
my @measure_array;
my $prev_measure = 1;
my $prev_gap = 1;
my $prev_forecast;
my $prev_gforecast;
my $prev_activecnt;
my $prev_incumbent;
my @gap_array;
my $first_gap;
my $start_time;
my $total_time;
my $next_time = 0;
my $incumbent_cnt = 0;
my $integer_control = 0;
my $base;
my $scale = 1;


# root node's left child and right child
my $leftChild;
my $rightChild;

# estimation definitions
my @forecast_array;
my @forecast_time;
my @b_t;
my @S_t;
my $alpha = 1;
my $gamma = 0.2;
my $beta = 0.8;

# estimation using MIP gap 
my @gforecast_array;
my @gb_t;
my @gS_t;

#######################    (1) INITIAL PROCESSING      ##############################

# open the input file
#open(INFILE, "<$infile") or die "Couldn't open file $infile\n";
#
# count the number of total & process lines in the tree output file
#GOLINE: while ($input = <INFILE>)
#{
#    # ignore comments
#    next GOLINE if ($input =~ /\#/);
#
#    $linecounter++;
#
#    # remove the end of line character
#    chomp($input);
#
#    # split the input line from spaces
#    my @values = split(' ', $input);
#
#    if(defined($values[1]))
#    {   
#        if($values[1] eq "heuristic")
#        {
#	      if(!defined($maxobj))
#	      {
#		   $maxobj = $values[2]; $minobj = $values[2];
#	      }
#
#	      if($values[2] > $maxobj)  {$maxobj = $values[2];}
#	      if($values[2] < $minobj)  {$minobj = $values[2];}
#       }
#    }
#        	
#    if(defined($values[5]))
#    {
#        # record extreme values of LP bound, iicount and iisum
#	  if(!defined($maxobj))
#	  {
#		$maxobj = $values[5]; 
#            $minobj = $values[5];
#	  }
#
#	  if($values[5] > $maxobj) 	 {$maxobj = $values[5];}
#	  if($values[5] < $minobj) 	 {$minobj = $values[5];}
#   }
#
#    if(defined($values[6]))
#    {
#	  if($values[6] > $maxiicount) {$maxiicount = $values[6];}
#    }
#
#    if(defined($values[7]))
#    {
#	  if($values[7] > $maxiisum) 	 {$maxiisum = $values[7];}
#    }
#
#    # get the total time
#    if(eof(INFILE))
#    {
#        print "total time: $values[0]\n";
#        $total_time = $values[0];
#    } 
#}
#close(INFILE);


# open the input file
open(INFILE, "<$infile") or die "Couldn't open file $infile\n";

# Record information about every image
# and extreme values of integer infeasibility and obj values

# report progress
print "Reading file and doing initial bookkeeping\n";

LINE: while ($input = <INFILE>)
{
    # ignore comments
    next LINE if ($input =~ /\#/);

    # remove the end of line character
    chomp($input);
    
    # split the input line from spaces
    my @values = split(' ', $input);
    
    # get optimization direction	
    if ($values[0] eq "d") { $opt_dir = $values[1]; next LINE; }

    # store the total time
    if(eof(INFILE)) {$total_time = $values[0];} 

    if ($next_time <= $timelimit || $timelimit < 0)
    {
	# store and report the start time
	unless(defined($start_time))
	{
	    $start_time = $values[0];
	    $next_time = $start_time + $interval;
	    print "start:$start_time\n";
	}

	# generate pictures if you read enough lines
	if(($values[0] > $next_time) && ($values[1] ne "fathomed") && ($values[1] ne "pregnant") && ($values[1] ne "heuristic") && ($values[1] ne "integer") && ($values[1] ne "candidate") && defined($rootID)&& ($integer_control == 0))
	{
	    # When calling subroutines, send them the current time as well
	    my $time_now = $values[0];
	    if($do_histogram)   {generate_histograms($imagename,$time_now);}
	    if($do_scatterplot) {generate_scatterplots($imagename,$time_now);}
	    if($do_tree)        {generate_trees($imagename,$time_now);}
	    if($do_fixedtree)   {generate_trees($imagename,$time_now,"fixed");}
	    calculate_measure(0,$values[0]);

	    $imagename++;
	    $next_time = $next_time + $interval;
	}

	# generate pictures if you just found a new incumbent
	# if integer control is greater than zero
	# that you've just found a new incumbent
	if(($values[1] ne "fathomed") && ($values[1] ne "pregnant") && ($values[1] ne "heuristic") && ($values[1] ne "integer") && ($values[1] ne "candidate") && ($integer_control > 0) && defined($rootID))
	{
	    # When calling subroutines, send them the current time as well
	    my $time_now = $values[0];
	    if($do_histogram)   {generate_histograms($imagename,$time_now);}
	    if($do_scatterplot) {generate_scatterplots($imagename,$time_now);}
	    if($do_tree)        {generate_trees($imagename,$time_now);}
	    if($do_fixedtree)   {generate_trees($imagename,$time_now,"fixed");}

	    calculate_measure(1,$values[0]);

	    # if integer control is equal to 1
	    # that means the new incumbent
	    # is not a heuristic incumbent and so, it has a parent.
	    if($integer_control == 1 && $do_path) {generate_incumbentpath($imagename, $time_now);}

	    $imagename++;
	    $next_time = $values[0] + $interval;

	    # reset the integer control
	    $integer_control = 0;
	}

	# increment the line counter
	$nlines++; 

	# report progress
	#if ($nlines % 1000 == 0) { print "$nlines processed ...\n"; }

	# get heuristic incumbent
	if($values[1] eq "heuristic") 
	{ 
	    # this sanity check is only valid for min problems.
	    # we don't need it at all if we are sure that the tree output is correct
	    if (defined($incumbent))
	    {
		if ($values[2] >= $incumbent) {next LINE;}
	    }

	    # shows that you just found a new heuristic incumbent
	    if(defined($rootID)){$integer_control = 2;}

	    # increment the total number of incumbents found so far
	    $incumbent_cnt++;
            
	    # assign the new incumbent value
	    $past_incumbent = $incumbent;
	    $incumbent= $values[2];

	    # record extreme values of LP bound
	    if(!defined($maxobj))
	    {
		$maxobj = $incumbent; 
		$minobj = $incumbent;
	    }

	    if($incumbent > $maxobj) {$maxobj = $incumbent;}
	    if($incumbent < $minobj) {$minobj = $incumbent;} 

	    next LINE;
	}
	
	if ($values[1] eq "candidate")
	{       
	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}
 
        # store information about that node
	  $nodes{$i}[0] =  $values[3];			# parent ID
	  $nodes{$i}[1] =  $values[4];  			# branch direction 

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}        

	  $nodes{$i}[2] = 0;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

	  # if optional lp bound information is provided
        # then set lp bound to this value
        # otherwise set it to its parent's lp bound

        if(defined($values[5]))  
        {
        	$nodes{$i}[3] = $values[5];

		# also check for the maximum and minimum objectives
	  	if($values[5] > $maxobj) 	 {$maxobj = $values[5];}
	  	if($values[5] < $minobj) 	 {$minobj = $values[5];}
	  }

	  if(!defined($values[5])) {$nodes{$i}[3] = $nodes{$values[3]}[3];}

	  # if optional iifeas cnt information is provided
        # then set iifeas cnt to this value
        # otherwise set it to its parent's iifeas cnt

        if(defined($values[6]))  
        {
        	$nodes{$i}[4] = $values[6];

            # also check for the maxiicnt
	      if($values[6] > $maxiicount) {$maxiicount = $values[6];}
        }

	  if(!defined($values[6])) {$nodes{$i}[4] = $nodes{$values[3]}[4];}

	  # if optional iifeas sum information is provided
        # then set iifeas sum to this value
        # otherwise set it to its parent's iifeas sum

        if(defined($values[7]))  
        {
        	$nodes{$i}[5] = $values[7];

		# also check for the maxiisum
	  	if($values[7] > $maxiisum) {$maxiisum = $values[7];}
        }

	  if(!defined($values[7])) {$nodes{$i}[5] = $nodes{$values[3]}[5];}

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;			 	      # total number of nodes branching out this node
        
        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        } 

        next LINE;
    }

    if ($values[1] eq "pregnant")
    {						
	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}

        # store information about that node
	  $nodes{$i}[0] = $values[3];			      # parent ID
	  $nodes{$i}[1] = $values[4];  			# branch direction

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}  

	  $nodes{$i}[2] = 1;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

	  $nodes{$i}[3] = $values[5];			    	# LP bound		      
	  $nodes{$i}[4] = $values[6];				# integer infeasibility count
	  $nodes{$i}[5] = $values[7];	   			# integer infeasibility sum

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;			 	      # total number of nodes branching out this node

	  # record extreme values of LP bound, iicount and iisum
	  if(!defined($maxobj))
	  {
	  	if (defined($incumbent)) {$maxobj = $incumbent; $minobj = $incumbent;}
		else 				 {$maxobj = $values[5]; $minobj = $values[5];}
	  }

	  if($values[5] > $maxobj) 	 {$maxobj = $values[5];}
	  if($values[5] < $minobj) 	 {$minobj = $values[5];}
	  if($values[6] > $maxiicount) {$maxiicount = $values[6];}
	  if($values[7] > $maxiisum) 	 {$maxiisum = $values[7];}

        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        } 

        next LINE;
    }

    if ($values[1] eq "branched")
    {				
	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}

        # store information about that node
	  $nodes{$i}[0] = $values[3];			      # parent ID
	  $nodes{$i}[1] = $values[4];  			# branch direction

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}  

	  $nodes{$i}[2] = 3;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

	  $nodes{$i}[3] = $values[5];			    	# LP bound		      
	  $nodes{$i}[4] = $values[6];				# integer infeasibility count
	  $nodes{$i}[5] = $values[7];	   			# integer infeasibility sum

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;			 	      # total number of nodes branching out this node

	  # record extreme values of LP bound, iicount and iisum
	  if(!defined($maxobj))
	  {
	  	if (defined($incumbent)) {$maxobj = $incumbent; $minobj = $incumbent;}
		else 				 {$maxobj = $values[5]; $minobj = $values[5];}
	  }

	  if($values[5] > $maxobj) 	 {$maxobj = $values[5];}
	  if($values[5] < $minobj) 	 {$minobj = $values[5];}
	  if($values[6] > $maxiicount) {$maxiicount = $values[6];}
	  if($values[7] > $maxiisum) 	 {$maxiisum = $values[7];}	

        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        }  

        next LINE;
    }

    if ($values[1] eq "integer")
    {	
        # shows that you just found a new incumbent
        $integer_control = 1;

        # increment the total number of incumbents found so far
        $incumbent_cnt++;

	  # store information about incumbent node
        $past_incumbent = $incumbent;
	  $incumbent = $values[5];
	  $incumbent_parent = $values[3];

	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}

        # store information about that node
	  $nodes{$i}[0] = $values[3];			      # parent ID
	  $nodes{$i}[1] = $values[4];  			# branch direction

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}  

	  $nodes{$i}[2] = 6;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

	  $nodes{$i}[3] = $values[5];			    	# LP bound		      
	  $nodes{$i}[4] = 0;					# integer infeasibility count
	  $nodes{$i}[5] = 0;	   				# integer infeasibility sum

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;			 	      # total number of nodes branching out this node

	  # record extreme values of LP bound
	  if(!defined($maxobj)) {$maxobj = $incumbent; $minobj = $incumbent;}

	  if($values[5] > $maxobj) 	 {$maxobj = $values[5];}
	  if($values[5] < $minobj) 	 {$minobj = $values[5];}

        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        }

        next LINE;
    }


    if ($values[1] eq "fathomed")
    {	
	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}
 
        # store information about that node
	  $nodes{$i}[0] = $values[3];			      # parent ID
	  $nodes{$i}[1] = $values[4];  			# branch direction

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}  

	  $nodes{$i}[2] = 5;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

        # if optional lp bound information is
        # provided then, set the lp bound
        # to the this value
        if(defined($values[5]))
        {
            $nodes{$i}[3] = $values[5];
        }
        
        # if optional value is not provided
        # either set it to the current incumbent
        # or set it to its parent's lp bound
        else
        {
            if($opt_dir == 120)	# minimization problem
            { 
        		if($nodes{$values[3]}[3] >= $incumbent)  {$nodes{$i}[3] = $nodes{$values[3]}[3];}
			else{$nodes{$i}[3] = $incumbent;}
            }

            if($opt_dir == 121)	# maximization problem
            {
        		if($nodes{$values[3]}[3] >= $incumbent)  {$nodes{$i}[3] = $incumbent;}
			else{$nodes{$i}[3] = $nodes{$values[3]}[3];}
            }
        }
		      
	  $nodes{$i}[4] = $nodes{$values[3]}[4];		# parent's integer infeasibility count
	  $nodes{$i}[5] = $nodes{$values[3]}[5];   	# parent's integer infeasibility sum

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;			 	      # total number of nodes branching out this node

        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        } 

        next LINE;			                 
    }

    if ($values[1] eq "infeasible")
    {	
	  # get the node ID
	  my $i = $values[2];

        # store the keys in chronological order
        push (@keylist,$i)    unless exists $nodes{$i};

	  # first node appered is always the root node
        # root's reference is always itself.
	  if(!defined($rootID)) {$rootID = $i; $nodes{$rootID}[11] = $rootID; $nodes{$rootID}[12] = 1;}
 
        # store information about that node
	  $nodes{$i}[0] = $values[3];			      # parent ID
	  $nodes{$i}[1] = $values[4];  			# branch direction

        if($nodes{$i}[0] == $rootID)
	  {
		if($nodes{$i}[1] eq "R") {$rightChild = $i;}
		if($nodes{$i}[1] eq "L") {$leftChild = $i;}
	  }

        # assign node's parent's left(right) child index
        # this is required to draw trees if not defined the 
        # node does not have a left(right) child yet.
	  if($nodes{$i}[1] eq "L") {$nodes{$values[3]}[10] = $i;}
        if($nodes{$i}[1] eq "R") {$nodes{$values[3]}[13] = $i;}  

	  $nodes{$i}[2] = 4;		  			# node status  0-candidate   1-pregnant  3-branched  
									# 4-infeasible  5-fathomed  6-integer

	  $nodes{$i}[3] = $nodes{$values[3]}[3];    	# parent's LP bound		      
	  $nodes{$i}[4] = $nodes{$values[3]}[4];		# parent's integer infeasibility count
	  $nodes{$i}[5] = $nodes{$values[3]}[5];   	# parent's integer infeasibility sum

	  $nodes{$i}[6] = 0;					# lower bound for the horizantal position
	  $nodes{$i}[7] = 0;					# upper bound for the horizantal position
	  $nodes{$i}[8] = 0;					# horizantal position
        $nodes{$i}[9] = 1;					# total number of nodes branching out this node

        # if node is not the root node set its ref_point
        # to its parent ref_point.
        if($i ne $rootID) 
        {
            unless(defined($nodes{$i}[11]))
            {
                $nodes{$i}[11] = $nodes{$values[3]}[11];
            }
        }
 
        next LINE;
    }

  }

}

close(INFILE);
print "Reading file finished ...\n";

# Make a final set of tree images
my $time_now = $total_time;
if($do_tree)      {generate_trees($imagename,$time_now);}
if($do_fixedtree) {generate_trees($imagename,$time_now,"fixed");}
$imagename++;

#################################################
# Process measures of progress and predictions

if ($do_predictions) 
{

    for(my $j = 1; $j<=$incumbent_cnt; $j++)
    {

	# create string for 3-digit representation of the number $incumbent_cnt
	my $j_string = sprintf("%03d", $j);

	my $measure_file = "ssg.$j_string";
	my $gap_file = "MIPgap.$j_string";

	if($#{$measure_array[$j]} > -1)
	{
		# Open data file for the the new measure value
		open(MEASUREDATA, "> $measure_file");
		foreach my $element (@{$measure_array[$j]})
		{
	    		print MEASUREDATA "$element\n";
		}
		close(MEASUREDATA);

		# open script file for the measure
		my $measure_script = "ssg$j_string.gnuplot";
		open(MEASURESCRIPT, "> $measure_script");

		# store the name of this file in plotting files
		print PLOTFILES "$measure_script\n";
		$plotfilenames[$#plotfilenames+1] = $measure_script;

		# set terminal for the output files
		print MEASURESCRIPT "set terminal png notransparent large\n\n";

		# make settings for the measure plot
		print MEASURESCRIPT "set output \"ssgplot$j_string.png\"\n";
		print MEASURESCRIPT "set title \"progress measure plot ($infile - $label)\"\n";
		print MEASURESCRIPT "set xlabel \"time\"\n";
		print MEASURESCRIPT "set ylabel \"measure\"\n";
		print MEASURESCRIPT "set autoscale\n";
		print MEASURESCRIPT "set yrange [0:1]\n";
		print MEASURESCRIPT "plot \'$measure_file\' with linespoints ti \"(SSG)\"\n";
		print MEASURESCRIPT "show output\n";
		close(MEASURESCRIPT);
	}

	else {print "ssgplot$j_string.png is not plotted since not enough observations were made between incumbents\n";}


	if($#{$gap_array[$j]} > -1)
	{
		# Open data file for the the gap value
		open(GAPDATA, "> $gap_file");
		foreach my $element (@{$gap_array[$j]})
		{
	    		print GAPDATA "$element\n";
		}
		close(GAPDATA);

		# open script file for the gap
		my $gap_script = "gap$j_string.gnuplot";
		open(GAPSCRIPT, "> $gap_script");
	
		# store the name of this file in plotting files
 		print PLOTFILES "$gap_script\n";
		$plotfilenames[$#plotfilenames+1] = $gap_script;

		# set terminal for the output files
		print GAPSCRIPT "set terminal png notransparent large\n\n";

		# make settings for the measure plot
		print GAPSCRIPT "set output \"gapplot$j_string.png\"\n";
		print GAPSCRIPT "set title \"MIP gap plot ($infile - $label)\"\n";
		print GAPSCRIPT "set xlabel \"time\"\n";
		print GAPSCRIPT "set ylabel \"MIP gap\"\n";
		print GAPSCRIPT "set autoscale\n";
		print GAPSCRIPT "set yrange [0:1]\n";
		print GAPSCRIPT "plot \'$gap_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\"\n";
		print GAPSCRIPT "show output\n";
		close(GAPSCRIPT);
	}

     	else {print "gapplot$j_string.png is not plotted since not enough observations were made between incumbents\n";}



	if(($#{$measure_array[$j]} > -1) && ($#{$gap_array[$j]} > -1))
      {
		# open script file for the gapmeasure
		my $gapmeasure_script = "gap_ssg$j_string.gnuplot";
		open(GAPMEASURESCRIPT, "> $gapmeasure_script");

	      # store the name of this file in plotting files
		print PLOTFILES "$gapmeasure_script\n";
		$plotfilenames[$#plotfilenames+1] = $gapmeasure_script;

		# set terminal for the output files
		print GAPMEASURESCRIPT "set terminal png notransparent large\n\n";

		# make settings for the measure plot
		print GAPMEASURESCRIPT "set output \"gap_ssg_plot$j_string.png\"\n";
		print GAPMEASURESCRIPT "set title \"SSG vs MIP gap ($infile - $label)\"\n";
		print GAPMEASURESCRIPT "set xlabel \"time\"\n";
		print GAPMEASURESCRIPT "set ylabel \"measure and MIP gap\"\n";
		print GAPMEASURESCRIPT "set autoscale\n";
		print GAPMEASURESCRIPT "set yrange [0:1]\n";
		print GAPMEASURESCRIPT "plot \'$measure_file\' with linespoints lt 3 ti \"(SSG)\", \'$gap_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\"\n";
		print GAPMEASURESCRIPT "show output\n";
		close(GAPMEASURESCRIPT);
	}

     	else {print "gap_ssg_plot$j_string.png is not plotted since not enough observations were made between incumbents\n";}


	# Open data file for the the last forecast values
	# if there is any forecast for the last incumbent
	if (($j == $incumbent_cnt) && ($#{$forecast_array[$j]} >= 0))
	{
	    my $forecast_file = "predictions(Prog.Measure)";
	    open(FORECASTDATA, "> $forecast_file");

	    foreach my $element (@{$forecast_array[$j]})
	    {
		print FORECASTDATA "$element\n";
	    }
	    close(FORECASTDATA);

	    my $gforecast_file = "predictions(MIP.gap)";
	    open(gFORECASTDATA, "> $gforecast_file");

	    foreach my $element (@{$gforecast_array[$j]})
	    {
		print gFORECASTDATA "$element\n";
	    }
	    close(gFORECASTDATA);

	    # Open data file for all forecast time values
	    my $forecasttime_file = "timelast";
	    open(FORECASTTIME, "> $forecasttime_file");

	    foreach my $element (@{$forecast_time[$j]})
	    {
		print FORECASTTIME "$element\n";
	    }

	    close(FORECASTTIME);

	    # open script file for the forecast
	    my $forecast_script = "forecast.gnuplot";
	    open(FORECASTSCRIPT, "> $forecast_script");

	    # store the name of this file in plotting files
	    print PLOTFILES "$forecast_script\n"; 
	    $plotfilenames[$#plotfilenames+1] = $forecast_script;

	    # set terminal for the output files
	    print FORECASTSCRIPT "set terminal png notransparent large\n\n";

	    # make settings for the measure plot
	    print FORECASTSCRIPT "set output \"forecast.png\"\n";
	    print FORECASTSCRIPT "set title \"prediction plot ($infile - $label)\"\n";
	    print FORECASTSCRIPT "set format x \"\"\n";
	    print FORECASTSCRIPT "set xlabel \"time\"\n";
	    print FORECASTSCRIPT "set ylabel \"prediction\"\n";
	    print FORECASTSCRIPT "set autoscale\n";
	    print FORECASTSCRIPT "plot \'$forecast_file\' with linespoints lt 3 ti \"(SSG)\", \'$gforecast_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\", $total_time lt 2 noti\n";
	    print FORECASTSCRIPT "show output\n";
	    close(FORECASTSCRIPT);
	}

      if (($j == $incumbent_cnt) && ($#{$forecast_array[$j]} < 0)) {print "Any plot is not generated for the last incumbent since not enough observations were made before termination\n";}


    }

    my $measure_file = "ProgMeasure.dat";
    my $gap_file = "MIPgap.dat";

    if ($#measure_array > -1)
    {
    	# Open data file for the all measure value
      open(MEASUREDATA, "> $measure_file");

    	for(my $j = 1; $j<=$incumbent_cnt; $j++)
    	{
		foreach my $element (@{$measure_array[$j]})
		{
	    		print MEASUREDATA "$element\n";
		}
    	}
    	close(MEASUREDATA);

    	# open script file for the measure
    	my $measure_script = "measure.gnuplot";
    	open(MEASURESCRIPT, "> $measure_script");

	# store the name of this file in plotting files
      print PLOTFILES "$measure_script\n"; 
      $plotfilenames[$#plotfilenames+1] = $measure_script;

    	# set terminal for the output files
    	print MEASURESCRIPT "set terminal png notransparent large\n\n";

      # make settings for the measure plot
    	print MEASURESCRIPT "set output \"measureplot.png\"\n";
    	print MEASURESCRIPT "set title \"progress measure plot ($infile - $label)\"\n";
    	print MEASURESCRIPT "set xlabel \"time\"\n";
    	print MEASURESCRIPT "set ylabel \"measure\"\n";
    	print MEASURESCRIPT "set autoscale\n";
    	print MEASURESCRIPT "set yrange [0:1]\n";
    	print MEASURESCRIPT "plot \'$measure_file\' with linespoints ti \"(SSG)\"\n";
    	print MEASURESCRIPT "show output\n";
    	close(MEASURESCRIPT);
    }

    else {print "measureplot.png is not plotted since not any measure calculated\n";}


    if($#gap_array > -1)
    {
    	# Open data file for the all gap value
    	open(GAPDATA, "> $gap_file");

    	for(my $j = 1; $j<=$incumbent_cnt; $j++)
    	{
		foreach my $element (@{$gap_array[$j]})
		{
	    		print GAPDATA "$element\n";
		}
    	}
    	close(GAPDATA);

    	# open script file for the gap
    	my $gap_script = "gap.gnuplot";
    	open(GAPSCRIPT, "> $gap_script");

    	print PLOTFILES "$gap_script\n";
	$plotfilenames[$#plotfilenames+1] = $gap_script;

    	# set terminal for the output files
    	print GAPSCRIPT "set terminal png notransparent large\n\n";

    	# make settings for the measure plot
    	print GAPSCRIPT "set output \"gapplot.png\"\n";
    	print GAPSCRIPT "set title \"MIP gap plot ($infile - $label)\"\n";
    	print GAPSCRIPT "set xlabel \"time\"\n";
    	print GAPSCRIPT "set ylabel \"MIP gap\"\n";
    	print GAPSCRIPT "set autoscale\n";
    	print GAPSCRIPT "set yrange [0:1]\n";
    	print GAPSCRIPT "plot \'$gap_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\"\n";
    	print GAPSCRIPT "show output\n";
    	close(GAPSCRIPT);
    }

    else {print "gapplot.png is not plotted since not any MIPgap calculated\n";}


    if(($#measure_array > -1) && ($#gap_array > -1))
    {
    	# open script file for the gapmeasure
    	my $gapmeasure_script = "gapmeasure.gnuplot";
    	open(GAPMEASURESCRIPT, "> $gapmeasure_script");

    	print PLOTFILES "$gapmeasure_script\n";
	$plotfilenames[$#plotfilenames+1] = $gapmeasure_script;

    	# set terminal for the output files
    	print GAPMEASURESCRIPT "set terminal png notransparent large\n\n";

    	# make settings for the measure plot
    	print GAPMEASURESCRIPT "set output \"gapmeasureplot.png\"\n";
    	print GAPMEASURESCRIPT "set title \"SSG vs MIP gap ($infile - $label)\"\n";
    	print GAPMEASURESCRIPT "set xlabel \"time\"\n";
    	print GAPMEASURESCRIPT "set ylabel \"measure and MIP gap\"\n";
    	print GAPMEASURESCRIPT "set autoscale\n";
    	print GAPMEASURESCRIPT "set yrange [0:1]\n";
    	print GAPMEASURESCRIPT "plot \'$measure_file\' with linespoints lt 3 ti \"(SSG)\", \'$gap_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\"\n";
    	print GAPMEASURESCRIPT "show output\n";
    	close(GAPMEASURESCRIPT);
    }

    else {print "gapmeasureplot.png is not plotted since not any MIPgap or ProgMeasure calculated\n";}


    if($#forecast_array > -1)
    {
    	# Open data file for all forecast values
    	my $forecast_file = "predictions(ProgMeasure)";
    	open(FORECASTDATA, "> $forecast_file");

    	my $forecast_counter = 0;

    	for(my $j = 0; $j<=$#forecast_array; $j++)
    	{
		foreach my $element (@{$forecast_array[$j]})
		{
	    		print FORECASTDATA "$element\n";
	    		$forecast_counter++;
		}
    	}
    	close(FORECASTDATA);

    	# Open data file for all gforecast values
    	my $gforecast_file = "predictions(MIPgap)";
    	open(gFORECASTDATA, "> $gforecast_file");

    	for(my $j = 0; $j<=$#gforecast_array; $j++)
    	{
		foreach my $element (@{$gforecast_array[$j]})
		{
	    		print gFORECASTDATA "$element\n";
		}
    	}
    	close(gFORECASTDATA);

    	# open script file for the forecast
    	my $forecast_script = "forecastall.gnuplot";
    	open(FORECASTSCRIPT, "> $forecast_script");

    	print PLOTFILES "$forecast_script\n";
	$plotfilenames[$#plotfilenames+1] = $forecast_script;

    	# set terminal for the output files
    	print FORECASTSCRIPT "set terminal png notransparent large\n\n";

    	# make settings for the measure plot
    	print FORECASTSCRIPT "set output \"forecast_all.png\"\n";
    	print FORECASTSCRIPT "set multiplot\n";
    	print FORECASTSCRIPT "set title \"prediction plot ($infile - $label)\"\n";
    	print FORECASTSCRIPT "set format x \"\"\n";
    	print FORECASTSCRIPT "set xlabel \"time\"\n";
    	print FORECASTSCRIPT "set ylabel \"prediction\"\n";
    	print FORECASTSCRIPT "set autoscale\n";
    	print FORECASTSCRIPT "set arrow from 0,2*$total_time to $forecast_counter-1,$total_time nohead lt 1\n";
    	print FORECASTSCRIPT "set arrow from 0,0.5*$total_time to $forecast_counter-1,$total_time nohead lt 1\n";
    	print FORECASTSCRIPT "plot \'$forecast_file\' with linespoints lt 3 ti \"(SSG)\", \'$gforecast_file\' with linespoints lt 4 pt 19 ti \"(MIP gap)\", $total_time lt 2 noti\n";

    	print FORECASTSCRIPT "set nomultiplot\n";
    	print FORECASTSCRIPT "show output\n";
    	close(FORECASTSCRIPT);
    }

    else {print "forecast_all.png is not plotted since not any forecast value calculated\n";}


    # Open data file for all forecast time values
    my $forecasttime_file = "time";
    open(FORECASTTIME, "> $forecasttime_file");

    for(my $j = 0; $j<=$#forecast_time; $j++)
    {
	foreach my $element (@{$forecast_time[$j]})
	{
	    print FORECASTTIME "$element\n";
	}
    }
    close(FORECASTTIME);

} # end of predictions

close(PLOTFILES);
print "Now plotting all pictures ...\n";
#system("$directories/BAK_gnuplotter","scriptfiles.dat");
foreach (@plotfilenames) { system("gnuplot","$_"); }


###################################     (1)    ##############################################

# Routine to find the maximum of a set of numbers.

sub max
{
    my (@numbers);

    @numbers = @_;

    my ($min, $max);

    $min = $numbers[0];
    $max = $numbers[0];

    foreach my $i (@numbers)
    {
        if ($i > $max)
        {
            $max = $i;
        }
        elsif ($i < $min)
        {
            $min = $i;
        }
    }

    return $max;
}


sub generate_scatterplots
{
    # get the image number and time
    my $i = $_[0];
    my $time_now = int($_[1]);

    # create string for 3-digit representation of the number $i
    my $i_string = sprintf("%03d", $i);

    # open script file for scatter plots
    my $scriptfile = "scatterplot$i_string.gnuplot";
    open(SCRIPTFILE, "> $scriptfile");

    # store the name of this file in plotting files
    print PLOTFILES "$scriptfile\n"; 
    $plotfilenames[$#plotfilenames+1] = $scriptfile;

    # determine the scale limits of the scatterplot
    my $scaling_maxobj;
    my $scaling_minobj; 
    my $scale_shift = $maxobj - $minobj;

    if ($scale_shift > 0)
    {
    	$scaling_maxobj = $maxobj + 0.01*$scale_shift;
       	$scaling_minobj = $minobj - 0.01*$scale_shift;
    }

    else
    {
    	$scaling_maxobj = 1.01*$maxobj;
    	if ($maxobj < 0) {$scaling_maxobj = 0.99*$maxobj;}

    	$scaling_minobj = 0.99*$minobj;
    	if ($minobj < 0) {$scaling_minobj = 1.01*$minobj;} 
    }



    # set terminal for the output files
    print SCRIPTFILE "set terminal png notransparent large\n\n";

    # make settings for the scatter plot
    print SCRIPTFILE "set output \"scatterplot.$i_string.png\"\n";
    print SCRIPTFILE "set title \"scatter plot ($infile ${time_now}s $label)\"\n";
    print SCRIPTFILE "set pointsize 0.8\n";
    print SCRIPTFILE "set nokey\n";
    print SCRIPTFILE "set xlabel \"sum of int. infeas.\"\n";
    print SCRIPTFILE "set ylabel \"obj. value\"\n";
    print SCRIPTFILE "set xrange [0:$maxiisum+2]\n";
    print SCRIPTFILE "set yrange [$scaling_minobj:$scaling_maxobj]\n";
    print SCRIPTFILE "set multiplot\n";

    # open output files for this image
    my $outfile= "scatterplot$i_string.dat";
    open (OUTFILE, "> $outfile") or die "Couldn't open file $outfile\n";

    # find the pregnant and candidate nodes
    # and print their parent's iifeas sum and lp bound information

    foreach my $j (keys %nodes)
    {
	if($nodes{$j}[2]==0 || $nodes{$j}[2]==1)
	{
	    # print iifeas sum and LP bound information
	    # to the output file. We will use this output file 
	    # when we are generating the scatter plots.

            # Do an optional check to omit fathomed nodes that have
	    # not been marked as fathomed by the solver.
	    if (!$check_fathom || 
		(defined($incumbent) && $nodes{$j}[3] <= $incumbent) ||
                !defined($incumbent) )
	    {  print OUTFILE "$nodes{$j}[5] $nodes{$j}[3]\n"; }
	}
    }
    close(OUTFILE);

    # plot the output file
    print SCRIPTFILE "plot \'$outfile\' w p pt 2 lt 1\n";

    # plot incumbent line
    if(defined($incumbent))
    {
	print SCRIPTFILE "plot $incumbent lt 2 lw 0.5\n";
    }

    # plot the incumbent parent information
    # if we have a heuristic incumbent it's
    # parent information is defined yet. for that
    # reason we need seperate if clauses for each case.

    if(defined($incumbent_parent))
    {
	print SCRIPTFILE "plot \"< echo $nodes{$incumbent_parent}[5] $nodes{$incumbent_parent}[3]\" w p pt 2 lt 2\n";
    }

    print SCRIPTFILE "set nomultiplot\n";
    print SCRIPTFILE "show output\n";

    close(SCRIPTFILE);

}

sub generate_histograms
{
    # get the image number and time
    my $i = $_[0];
    my $time_now = int($_[1]);

    # set the number of bins
    my $bins = 20;

    # create string for 3-digit representation of the number $i
    my $i_string = sprintf("%03d", $i);

    # Open script file for histograms
    my $scriptfile = "histogram$i_string.gnuplot";
    open(SCRIPTFILE, "> $scriptfile");

    # store the name of this file in plotting files
    print PLOTFILES "$scriptfile\n"; 
    $plotfilenames[$#plotfilenames+1] = $scriptfile;

    # set terminal for the output files
    print SCRIPTFILE "set terminal png notransparent large\n\n";

    # make settings for the histogram
    print SCRIPTFILE "set output \"histogram$i_string.png\"\n";
    print SCRIPTFILE "set title \"histogram ($infile ${time_now}s $label)\"\n";
    print SCRIPTFILE "set xlabel \"obj. value\"\n";
    print SCRIPTFILE "set ylabel \"number of nodes\"\n";
    print SCRIPTFILE "set nokey\n";
    print SCRIPTFILE "set ticscale 0.001\n";

    # define an array to store the
    # lp bounds of parent's of pregnant 
    # and candidate nodes
    my @obj_array;

    # find the pregnant and candidate nodes
    # and store their lp bound information

    foreach my $j (keys %nodes)
    {
	if($nodes{$j}[2]==0 || $nodes{$j}[2]==1)
	{
            # Do an optional check to omit fathomed nodes that have
	    # not been marked as fathomed by the solver.
	    if (!$check_fathom || 
		(defined($incumbent) && $nodes{$j}[3] <= $incumbent) ||
                !defined($incumbent) )
	    {  push (@obj_array,$nodes{$j}[3]); }
	}
    }

    if ($#obj_array >= 0)
    {
	# sort @obj_array in ascending order.
	(@obj_array) = sort{$a<=>$b}(@obj_array);

        # set overall lp bound for this picture 
	my $lp_bound = $obj_array[0];

	# set binsize
	my $binsize = ($maxobj - $minobj)/$bins;
        if(($binsize <= EPS) && ($binsize >= -1*EPS))
        {
        	$bins = 1;
        }

	# define some important arrays and variables
	my @histogram;
        my @lower_bound;
        my @upper_bound;
	
        my $incumbent_xcoord;
        my $incumbent_bin;
        
        my $lpbound_xcoord;
        my $lpbound_bin;

        my $maxbin = 0;
	
        # initialize the histogram array
        for(my $k = 0; $k <$bins; $k++) {$histogram[$k] = 0;}

	# divide the @obj_array into bins
	for(my $k = 0; $k <$bins; $k++)
	{
	    # calculate lower & upper bounds of the bin
	    $lower_bound[$k] = $minobj+$k*$binsize;
	    $upper_bound[$k] = $minobj+($k+1)*$binsize;

	    # find the interval of the incumbent
	    if(defined($incumbent))
	    {
		if(($incumbent - $upper_bound[$k] <= EPS) && ($incumbent - $lower_bound[$k] >= -1*EPS))
		{
		    # calculate the position of the incumbent line in the graph
                    if(($binsize <= EPS) && ($binsize >= -1*EPS))
		    {
			$incumbent_xcoord = 1;
			$incumbent_bin = 1;
		    }

                    else
		    {
		    	$incumbent_xcoord = ($incumbent - $lower_bound[$k])/$binsize + ($k+0.5);
		        $incumbent_bin = $k+1;
	            }  
		}
	    }

	    # find the interval of the lp bound
	    if(($lp_bound - $upper_bound[$k] <= EPS) && ($lp_bound - $lower_bound[$k] >= -1*EPS))
	    {
		# calculate the position of the lp bound line in the graph
		if(($binsize <= EPS) && ($binsize >= -1*EPS))
		{
			$lpbound_xcoord = 0;
			$lpbound_bin = 1;
		}

		else
		{
			$lpbound_xcoord = ($lp_bound - $lower_bound[$k])/$binsize + ($k+0.5);
			$lpbound_bin = $k+1;
		}	
	    }
	    
	    # find the interval of each element of the obj_array
	    foreach my $element(@obj_array)
	    {
		if(($element - $upper_bound[$k] <= EPS) && ($element - $lower_bound[$k] >= -1*EPS))
		{
		    $histogram[$k]++;
		}
	    }
	}
	
	# open an output file for the histogram
	open(HISTOUT, ">histogram$i_string.dat");

	for(my $k = 1; $k <=$bins; $k++)
	{
	    # print the histogram information to the output file
	    print HISTOUT "$k $histogram[$k-1]\n";

	    # keep track of the maximum bin
	    # we need this information to set
	    # the yrange of the histogram
	    if($histogram[$k-1]>$maxbin) {$maxbin = $histogram[$k-1];}
	}
	close(HISTOUT);

	# set the xrange and the yrange of the histogram
	print SCRIPTFILE "set xrange [0:$bins+1]\n";
	print SCRIPTFILE "set yrange [0:$maxbin*1.2]\n";

	# round lp_bound and incumbent up to three digits
	# this is for printing purposes only
	my $roundedlp = sprintf("%.3f", $lp_bound);
	
	# set the labels of the x-axis
	print SCRIPTFILE "set xtics (";

	if(defined($incumbent))
	{    
	    my $roundedinc = sprintf("%.3f", $incumbent);
	    print SCRIPTFILE "\"$roundedlp\"$lpbound_bin, ";
	    print SCRIPTFILE "\"$roundedinc\"$incumbent_bin)\n";
	}

	if(!defined($incumbent))
	{
	    print SCRIPTFILE "\"$roundedlp\"$lpbound_bin)\n";
	}

	# set multiplot
	print SCRIPTFILE "set multiplot\n";

	# draw the histogram
	print SCRIPTFILE "plot \'histogram$i_string.dat\' with boxes fs solid 0.2\n";

	# finally draw the vertical lp_bound 
	# and incumbent lines
	print SCRIPTFILE "set parametric\n";
	print SCRIPTFILE "set trange [0:$maxbin*1.5]\n";
	print SCRIPTFILE "plot $lpbound_xcoord,t lt 2\n";
	if(defined($incumbent))
	{
	    print SCRIPTFILE "plot $incumbent_xcoord,t lt 5\n";
	}
	print SCRIPTFILE "unset parametric\n";

	print SCRIPTFILE "set nomultiplot\n";
	print SCRIPTFILE "show output\n";

	close(SCRIPTFILE);

    }

}

sub generate_incumbentpath
{
    if(defined($incumbent_parent))
    {
	# get the image number and time
	my $i = $_[0];
	my $time_now = int($_[1]);

        # create string for 3-digit representation of the number $i
        my $i_string = sprintf("%03d", $i);

        # Open script file for incumbentpath
        my $scriptfile = "incumbentpath$i_string.gnuplot";
        open(SCRIPTFILE, "> $scriptfile");

        # store the name of this file in plotting files
        print PLOTFILES "$scriptfile\n"; 
	$plotfilenames[$#plotfilenames+1] = $scriptfile;
	

    	# determine the scale limits of the incumbent path plot
    	my $scaling_maxobj;
    	my $scaling_minobj; 
    	my $scale_shift = $maxobj - $minobj;

    	if ($scale_shift > 0)
    	{
    		$scaling_maxobj = $maxobj + 0.01*$scale_shift;
        	$scaling_minobj = $minobj - 0.01*$scale_shift;
    	}

	else
    	{
    		$scaling_maxobj = 1.01*$maxobj;
    		if ($maxobj < 0) {$scaling_maxobj = 0.99*$maxobj;}

    		$scaling_minobj = 0.99*$minobj;
    		if ($minobj < 0) {$scaling_minobj = 1.01*$minobj;} 
    	}


        # set terminal for the output files
        print SCRIPTFILE "set terminal png notransparent large\n\n";

        # make settings for the incumbent path
        print SCRIPTFILE "set title \"incumbent path ($infile ${time_now}s $label)\"\n";
        print SCRIPTFILE "set output \"incumbentpath$i_string.png\"\n";
        print SCRIPTFILE "set xlabel \"sum of int. infeas.\"\n";
        print SCRIPTFILE "set xtics auto\n";
        print SCRIPTFILE "set pointsize 0.8\n";
        print SCRIPTFILE "set nokey\n";
        print SCRIPTFILE "set multiplot\n";
        print SCRIPTFILE "set ylabel \"obj. value\"\n";
        print SCRIPTFILE "set xrange [0:$maxiisum+2]\n";  
    	print SCRIPTFILE "set yrange [$scaling_minobj:$scaling_maxobj]\n\n";


        # open an output file for the incumbent path
        open(INCUMBENTOUT, ">incumbent$i_string.dat");

        # first print the incumbent's information
	print INCUMBENTOUT "0 $incumbent\n";

        # print the iifeas sum and lp bound information
        # until you reach the root node
        # we shouldn't use any ID information while traversing 
        # the tree because we assume that ID of a can be anything
        # as far as it is unique

        # get the incumbent parent ID
        my $parent = $incumbent_parent;

        while($parent ne $nodes{$rootID}[0])
        {
	    print INCUMBENTOUT "$nodes{$parent}[5] $nodes{$parent}[3]\n";
	    $parent = $nodes{$parent}[0];
        }
        close(INCUMBENTOUT);
	
        # finally draw the graph and show the output
        print SCRIPTFILE "plot \'incumbent$i_string.dat\' with lines lt 2\n";
        print SCRIPTFILE "plot \'incumbent$i_string.dat\' w p pt 2\n";

        print SCRIPTFILE "set nomultiplot\n";
        print SCRIPTFILE "show output\n\n";

	close(SCRIPTFILE);
    }

}

sub generate_trees
{
    # get the image number and time
    my $i = $_[0];
    my $time_now = int($_[1]);

    # The third argument, if present, indicates fixed horizontal positions
    my $fixed = 0;
    if (defined($_[2]) && $_[2] eq "fixed")  { $fixed = 1; }

    # create string for 3-digit representation of the number $i
    my $i_string = sprintf("%03d", $i);

    # Set the prefix for files
    my $prefix = "tree";
    if ($fixed == 1) { $prefix = "tree_alt"; }

    # Open script file for the tree
    my $scriptfile = "$prefix$i_string.gnuplot";
    open(TREESCRIPT, "> $scriptfile");

    # store the name of this file in plotting files
    print PLOTFILES "$scriptfile\n"; 
    $plotfilenames[$#plotfilenames+1] = $scriptfile;

    # determine the scale limits of the tree
    my $scaling_maxobj;
    my $scaling_minobj; 
    my $scale_shift = $maxobj - $nodes{$rootID}[3];

    if ($scale_shift > 0)
    {
    	$scaling_maxobj = $maxobj + 0.01*$scale_shift;
        $scaling_minobj = $nodes{$rootID}[3] - 0.01*$scale_shift;
    }

    else
    {
    	$scaling_maxobj = 1.01*$maxobj;
    	if ($maxobj < 0) {$scaling_maxobj = 0.99*$maxobj;}

    	$scaling_minobj = 0.99*$nodes{$rootID}[3];
    	if ($nodes{$rootID}[3] < 0) {$scaling_minobj = 1.01*$nodes{$rootID}[3];} 
    }

    # make settings for the tree
    print TREESCRIPT "set terminal png notransparent large\n";
    print TREESCRIPT "set output \"$prefix$i_string.png\"\n";
    print TREESCRIPT "set multiplot\n";
    print TREESCRIPT "set nokey\n";
    print TREESCRIPT "set autoscale\n";
    print TREESCRIPT "set ticscale 0.001\n";
    print TREESCRIPT "set pointsize 0.5\n";
    print TREESCRIPT "set xrange [-0.1:1.1]\n";
    print TREESCRIPT "set yrange [$scaling_maxobj:$scaling_minobj]\n";
    print TREESCRIPT "set format x \"\"\n";
    print TREESCRIPT "set ylabel \"obj. value\"\n";
    print TREESCRIPT "set title \"B&B tree ($infile ${time_now}s $label)\"\n\n";

    # traverse the tree and calculate the subtree sizes
    # this is required to set the horizantal positions of nodes
    # we shouldn't use any ID information while traversing 
    # the tree because we assume that ID of a can be anything
    # as far as it is unique

    # reset all subtree size information
    foreach my $j (keys %nodes) {$nodes{$j}[9] = 1;}

    foreach my $j (keys %nodes)
    {
	my $parent = $nodes{$j}[0];
	while($parent ne $nodes{$rootID}[0])
	{
	    if(!defined($parent)) {print "parent of $j is not defined\n"; next DONE;}
	    $nodes{$parent}[9]++;
	    $parent = $nodes{$parent}[0];
	}
    } 

    # set the upper and lower bound of the root node's position
    $nodes{$rootID}[6] = 0;
    $nodes{$rootID}[7] = 1;
    #$nodes{$rootID}[9] = 1;
    #if(defined($leftChild)) {$nodes{$rootID}[9] = $nodes{$rootID}[9] + $nodes{$leftChild}[9];}
    #if(defined($rightChild)) {$nodes{$rootID}[9] = $nodes{$rootID}[9] + $nodes{$rightChild}[9];}

    if ($fixed == 0)
    {
	if(defined($leftChild))
	{
	    # set the horizontal position of the root node
	    $nodes{$rootID}[8] = $nodes{$leftChild}[9]/$nodes{$rootID}[9]; 
	}
	
	else
	{
	    if(defined($rightChild)) 
	    {
		$nodes{$rootID}[8] = 1 - ($nodes{$rightChild}[9]/$nodes{$rootID}[9]);
	    }
	    else {$nodes{$rootID}[8] = 0.5;}
	}
    }
    else  # fixed horizontal positions
    {
	$nodes{$rootID}[8] = 0.5;
    }
    
    # traverse the tree and set the horizontal position for each node
    # first element of the @keylist array must always be
    # the root node, so that we can start assigining lower and
    # upper bounds from the second element of the list

    for(my $j = 1; $j<=$#keylist; $j++)
    {
        # get the node ID
        my $s = $keylist[$j];

	# get the parent of that node
	my $parent = $nodes{$s}[0];

        # right branch
	if($nodes{$s}[1] eq "R")
        {
	    # assign the lower and upper bound for the horizontal position
	    $nodes{$s}[6] = $nodes{$parent}[8];					
	    $nodes{$s}[7] = $nodes{$parent}[7];

	    my $range = $nodes{$s}[7] - $nodes{$s}[6];
	    my $ratio;
	    if (!defined($nodes{$s}[10])) {$ratio = 0.5;}		      # no left child
	    else {$ratio = $nodes{$nodes{$s}[10]}[9]/$nodes{$s}[9];}     # there is a left child

	    # assign the horizontal position 
           if ($fixed == 0)
           { $nodes{$s}[8] = $nodes{$s}[6] + $range*$ratio; }
           else  # fixed horizontal positions
           { $nodes{$s}[8] = ($nodes{$s}[6] + $nodes{$s}[7])/2; }
        }

        # left branch
	if($nodes{$s}[1] eq "L")
        {
	    # assign the lower and upper bound for the horizontal position
	    $nodes{$s}[6] = $nodes{$parent}[6];
	    $nodes{$s}[7] = $nodes{$parent}[8];	

	    my $range = $nodes{$s}[7] - $nodes{$s}[6];
	    my $ratio;
	    if (!defined($nodes{$s}[10])) {$ratio = 0.5;}		      # no left child
	    else {$ratio = $nodes{$nodes{$s}[10]}[9]/$nodes{$s}[9];}     # there is a left child

	    # assign the horizontal position 
	    if ($fixed == 0)
	    { $nodes{$s}[8] = $nodes{$s}[6] + $range*$ratio; }
	    else  # fixed horizontal positions
	    { $nodes{$s}[8] = ($nodes{$s}[6] + $nodes{$s}[7])/2; }
        }
    }

    # open output files for the tree
    open (TREECANDIDATE, "> ${prefix}_candidate$i_string.dat") or die "Couldn't open file candidate$i_string.dat\n";
    open (TREEPREGNANT, "> ${prefix}_pregnant$i_string.dat") or die "Couldn't open file pregnant$i_string.dat\n";
    open (TREEBRANCHED, "> ${prefix}_branched$i_string.dat") or die "Couldn't open file branched$i_string.dat\n";
    open (TREEINFEASIBLE, "> ${prefix}_infeasible$i_string.dat") or die "Couldn't open file infeasible$i_string.dat\n";
    open (TREEFATHOMED, "> ${prefix}_fathomed$i_string.dat") or die "Couldn't open file fathomed$i_string.dat\n";
    open (TREEINTEGER, "> ${prefix}_integer$i_string.dat") or die "Couldn't open file integer$i_string.dat\n";

    # define and initialize counters
    my $candidate_cnt = 0;
    my $pregnant_cnt = 0;
    my $branched_cnt = 0;
    my $infeasible_cnt = 0;
    my $fathomed_cnt = 0;
    my $integer_cnt = 0;

    # traverse the tree and send output
    # to the related data file

    foreach my $j (keys %nodes)
    {
	if($nodes{$j}[2] == 0) 
	{ 
          # Do an optional check to omit fathomed nodes that have
	    # not been marked as fathomed by the solver.
	    if (!$check_fathom || 
		 (defined($incumbent) && $nodes{$j}[3] <= $incumbent) ||
             !defined($incumbent) )
		{
			print TREECANDIDATE  "$nodes{$j}[8] $nodes{$j}[3]\n"; $candidate_cnt++;
		}
	    else
		{
			print TREEFATHOMED   "$nodes{$j}[8] $nodes{$j}[3]\n"; $fathomed_cnt++;	
		}
	}

	if($nodes{$j}[2] == 1) 
	{ 
          # Do an optional check to omit fathomed nodes that have
	    # not been marked as fathomed by the solver.
	    if (!$check_fathom || 
		 (defined($incumbent) && $nodes{$j}[3] <= $incumbent) ||
             !defined($incumbent) )
		{
			print TREEPREGNANT   "$nodes{$j}[8] $nodes{$j}[3]\n"; $pregnant_cnt++;
		}
	    else
		{
			print TREEFATHOMED   "$nodes{$j}[8] $nodes{$j}[3]\n"; $fathomed_cnt++;	
		}
	}

	if($nodes{$j}[2] == 3) { print TREEBRANCHED   "$nodes{$j}[8] $nodes{$j}[3]\n"; $branched_cnt++;}
	if($nodes{$j}[2] == 4) { print TREEINFEASIBLE "$nodes{$j}[8] $nodes{$j}[3]\n"; $infeasible_cnt++;}
	if($nodes{$j}[2] == 5) { print TREEFATHOMED   "$nodes{$j}[8] $nodes{$j}[3]\n"; $fathomed_cnt++;}
	if($nodes{$j}[2] == 6) { print TREEINTEGER    "$nodes{$j}[8] $nodes{$j}[3]\n"; $integer_cnt++;}

	# don't draw upward branches
	# if it is a root node
	if($j ne $rootID)
	{
	    my $parent = $nodes{$j}[0];
	    print TREESCRIPT "set arrow from $nodes{$parent}[8],$nodes{$parent}[3] to $nodes{$j}[8],$nodes{$j}[3] nohead lt -1 lw 0.2\n";
	}
    }

    close(TREECANDIDATE);
    close(TREEPREGNANT);
    close(TREEBRANCHED);
    close(TREEINFEASIBLE);
    close(TREEFATHOMED);
    close(TREEINTEGER);

    # first plot the root node
    print TREESCRIPT "plot \"< echo $nodes{$rootID}[8] $nodes{$rootID}[3]\" w p lt 2 pt 7\n";
    # After the first plot, turn off the arrows or they will be drawn again
    print TREESCRIPT "unset arrow\n";
    
    if($branched_cnt)   {print TREESCRIPT "plot \'${prefix}_branched$i_string.dat\' w p lt 2 pt 7\n";}
    if($fathomed_cnt)   {print TREESCRIPT "plot \'${prefix}_fathomed$i_string.dat\' w p lt 5 pt 7\n";}
    if($infeasible_cnt) {print TREESCRIPT "plot \'${prefix}_infeasible$i_string.dat\' w p lt 3 pt 7\n";}
    if($pregnant_cnt)   {print TREESCRIPT "plot \'${prefix}_pregnant$i_string.dat\' w p lt 7 pt 7\n";}
    if($candidate_cnt) {print TREESCRIPT "plot \'${prefix}_candidate$i_string.dat\' w p lt 6 pt 7\n";}
    if($integer_cnt)    {print TREESCRIPT "plot \'${prefix}_integer$i_string.dat\' w p lt 1 pt 7\n";}

    if(defined($incumbent)) {print TREESCRIPT "plot $incumbent lt 1 lw 0.5\n";}

    print TREESCRIPT "set nomultiplot\n";
    print TREESCRIPT "show output\n\n";
    close(TREESCRIPT);
}

sub calculate_measure
{
    # get the parameter
    # 1 -> found a new incumbent
    # 0 -> regular measure
    my $i = $_[0];
    my $time = $_[1];

    my $measure = 0;
    my $ssg = 0;
    my $activecnt = 0;
    my $counter = 0;
    my $lp_bound;
    my $id;
    my %minobj;
    my $F_t;
    my $gF_t;

    if (defined($incumbent))
    {
         foreach my $j (keys %nodes)
         {
              # if it is candidate or pregnant and not fathomed still 
              # caution: second check assumes minimization problem

	        if(($nodes{$j}[2]==0 || $nodes{$j}[2]==1) && ($nodes{$j}[3]<$incumbent))
	        {
			$activecnt++;

                   unless(exists $minobj{$nodes{$j}[11]}) {$minobj{$nodes{$j}[11]} = $nodes{$j}[3];}
                   if ($nodes{$j}[3] < $minobj{$nodes{$j}[11]}) {$minobj{$nodes{$j}[11]} = $nodes{$j}[3];}

                   # if you've just found an incumbent
                   if($i == 1)
                   { 
                       	# set the reference point to itself
                       	$nodes{$j}[11] = $j;

                        # calculate measure
                        $measure = $measure + ($incumbent - $nodes{$j}[3]); 

                        $counter++;
                   }

                   # find the overall lp-bound, assumes minimization problem
                   unless(defined($lp_bound)) {$lp_bound = $nodes{$j}[3];}
                   if($lp_bound > $nodes{$j}[3]) {$lp_bound = $nodes{$j}[3]; $id = $j;}
	       }
         }

         my $temp = 0;

         # find sum of subtree gaps
         foreach my $j (keys %minobj)
         {
                $ssg = $ssg + ($incumbent - $minobj{$j});
                my $inc = ($incumbent - $minobj{$j});
		    #print "i:$i    ssg:$ssg     inc:$inc    obj:$minobj{$j}\n";
		    $temp++;
         }

         # if you've just found an incumbent
         if($i == 1) 
         {
          
              # if you calculated measure for the previous incumbent
              if (defined($prev_incumbent)) 
              {
                   $ssg = 0;

                   foreach my $j (keys %minobj)
                   {
                         #print "i:$i    obj:$minobj{$j}   prev_incumbent:$prev_incumbent\n";
                         if($minobj{$j} < $prev_incumbent) {$ssg = $ssg + ($prev_incumbent - $minobj{$j});}
                   }

                   # update the scale factor
                   if($measure >= EPS)
                   {
                   	$scale = $scale*$ssg/$measure;
                   }

                   #print "i:$i  prev_inc:$prev_incumbent	inc:$incumbent	scale:$scale   ssg:$ssg   measure:$measure  subtree:$temp\n";
              }
         }

         # if you already have an incumbent
         if($i == 0)
         {
              $measure = $ssg;
         }

         # define the base
         unless (defined($base)) {$base = $measure; print "base: $base\n";}

         # store current gap information

         my $gap = 0;
         if($activecnt)
         {
         	$gap = ($incumbent - $lp_bound);
         }
         
         if (!defined($first_gap)) 
         {
          	 $first_gap = $gap;
         }
         
         $gap = $gap/($first_gap+EPS);
         push (@{$gap_array[$incumbent_cnt]}, $gap);

         # scale down the measure
         $measure = $measure*$scale/$base;

         # store the measure
         push (@{$measure_array[$incumbent_cnt]}, $measure);

	   # make estimation for the total completion time
         # if you have at least 4 measurements for the current incumbent.
         # we don't make any prediction at the very beginging because the decrease 
         # is always faster at the begining of the solution process
         # and this causes very small predictions at the begining

         # alpha: (btwn 0 and 1) weight given to the current point where (1-alpha) is 
         # the weight given to the value obtained by adding the past trend with the
         # point one before the current one. one could naively set alpha to 1 if (s)he
         # doesn't care about the history of the process. 
    
         # gamma: (btwn 0 and 1) weight given to the current trend where (1-gamma) is 
         # the weight given to the past trend. one could naively set alpha to 1 if (s)he
         # doesn't care about the history of the process. however, this is not recommended
         # since a gamma value close to zero helps preventing flactuating estimations and 
         # results in a smaller variation compared to a large one.
 
         # beta: (btwn 0 and 1) if beta is closer the zero then the prediction method would
         # more often overestimate the total solution time at the begining. This might be a
         # desirable property, since we dont't want estimations to increase as time passes.
         # one could naively set beta to 1 if (s)he doesn't care about increasing estimations.

         # set the initial values for the estimation method
         if($#{$measure_array[$incumbent_cnt]} > 2 && !defined($b_t[$incumbent_cnt]))
         {
              $b_t[$incumbent_cnt] = log(1+$measure_array[$incumbent_cnt][3]**$beta) - log(1+$measure_array[$incumbent_cnt][2]**$beta);
              $S_t[$incumbent_cnt] = log(1+$measure_array[$incumbent_cnt][2]**$beta);
         }

         # update the model parameters and calculate the estimate
         if(defined($b_t[$incumbent_cnt]))
         {
	     if(($measure - $prev_measure) <= -1*EPS)
           {
              my $temp = $S_t[$incumbent_cnt];
              $S_t[$incumbent_cnt] = $alpha*log(1+$measure**$beta) + (1-$alpha)*($S_t[$incumbent_cnt] + $b_t[$incumbent_cnt]);
              $b_t[$incumbent_cnt] = $gamma*($S_t[$incumbent_cnt] - $temp) + (1-$gamma)*$b_t[$incumbent_cnt];

              $F_t = (-$S_t[$incumbent_cnt]/$b_t[$incumbent_cnt])*$interval + $time;
              # $F_t = (-$S_t[$incumbent_cnt]/$b_t[$incumbent_cnt])*$interval;
           }

           else 
           {
             if(defined($prev_forecast)) 
                {
                 $F_t = ($prev_forecast-$time+$interval)*$activecnt/$prev_activecnt + $time;
                 # $F_t = ($prev_forecast-$time+$interval)*$activecnt/$prev_activecnt;
                }
             else
                {
                 $F_t = $activecnt*($time/($#keylist-$activecnt+1)) + $time;
                 # $F_t = $activecnt*($time/($#keylist-$activecnt+1)); 
                } 
           }

	     # store the forecast
           push (@{$forecast_array[$incumbent_cnt]}, $F_t);
         }

         # set the initial values for the estimation method
         if($#{$gap_array[$incumbent_cnt]} > 2 && !defined($gb_t[$incumbent_cnt]))
         {
              $gb_t[$incumbent_cnt] = log(1+$gap_array[$incumbent_cnt][3]**$beta) - log(1+$gap_array[$incumbent_cnt][2]**$beta);
              $gS_t[$incumbent_cnt] = log(1+$gap_array[$incumbent_cnt][2]**$beta);
         }

         # update the model parameters and calculate the estimate
         if(defined($gb_t[$incumbent_cnt]))
         {
	     if(($gap - $prev_gap) <= -1*EPS)
           {
              my $temp = $gS_t[$incumbent_cnt];
              $gS_t[$incumbent_cnt] = $alpha*log(1+$gap**$beta) + (1-$alpha)*($gS_t[$incumbent_cnt] + $gb_t[$incumbent_cnt]);
              $gb_t[$incumbent_cnt] = $gamma*($gS_t[$incumbent_cnt] - $temp) + (1-$gamma)*$gb_t[$incumbent_cnt];

              $gF_t = (-$gS_t[$incumbent_cnt]/$gb_t[$incumbent_cnt])*$interval + $time;
              # $gF_t = (-$gS_t[$incumbent_cnt]/$gb_t[$incumbent_cnt])*$interval;
           }

           else 
           {
             if(defined($prev_gforecast)) 
                {
                 $gF_t = ($prev_gforecast-$time+$interval)*$activecnt/$prev_activecnt + $time;
                 # $gF_t = ($prev_gforecast-$time+$interval)*$activecnt/$prev_activecnt;
                }
             else
                { 
                 $gF_t = $activecnt*($time/($#keylist-$activecnt+1)) + $time;
                 # $gF_t = $activecnt*($time/($#keylist-$activecnt+1));
                }
           }

	     # store the forecast
           push (@{$gforecast_array[$incumbent_cnt]}, $gF_t);

           # store the time of the predictions
           push (@{$forecast_time[$incumbent_cnt]}, $time);

         }

         # update the previous variables
         $prev_measure = $measure;
         $prev_gap = $gap;
         $prev_forecast = $F_t;
         $prev_gforecast = $gF_t;
         $prev_activecnt = $activecnt;
         $prev_incumbent = $incumbent;

        if(defined($F_t)) 
        {
             print ("measure($incumbent_cnt) = $measure  gap = $gap  forecast = $F_t  gap_forecast = $gF_t  num = $activecnt   time = $time\n");
        }
         	 
        else {print ("measure($incumbent_cnt) = $measure  num = $activecnt  time = $time\n");}

    }
     
}

sub print_usage 
{
    print "Usage: [perl] BAK_visual.pl [options] <input file>\n";
    print "Options:\n";
    print "  --help          print this help message\n";
    print "  --interval t    t is time in seconds between images\n";
    print "  --timelimit t   t is time limit ???\n";
    print "  --label s       s is a string to add to all image titles\n";
    print "  --fathom        do an internal check for fathomed nodes\n";
    print "  --nofathom      use the input file data with no check\n";
    print "\n";
    print "  --histogram     print histograms\n";
    print "  --nohistogram\n";
    print "  --scatterplot   print scatterplots\n";
    print "  --noscatterplot\n";
    print "  --path          print paths of incumbents (in scatterplot)\n";
    print "  --nopath\n";
    print "  --tree          print trees\n";
    print "  --notree\n";
    print "  --fixedtree     print trees with fixed horizontal positions\n";
    print "  --nofixedtree\n";
    print "  --predictions   print time predictions\n";
    print "  --nopredictions\n";
    print "  --all           print all images\n";
    print "\n";
}
